//Procedures for measuring the quality of solutions
//and verifying the correctness (optimality) of the algorithm

#include <stdio.h>
#include <math.h>
#include "ot.h"
#include "hyperorient.h"

//Recursive procedure used by DeriveInvPermutation to record the permutation
//given by an oriented dt into a lookup table
void WritePermutation(dt_node * dt,int * Pi,int * count)
{
  int orientation;

  //If this is a leaf, write it down
  if(dt->v != NOT_LEAF)
    {
      Pi[(*count)] = dt->v;
      (*count)++;
    }
  else
    {
      if(dt->orientation == -1) orientation = 0; else orientation = dt->orientation;

      //Otherwise, go to the left child, keeping in mind the orientation,
      //then go to the right child
      WritePermutation(dt->child[orientation], Pi, count);
      WritePermutation(dt->child[1 - orientation], Pi, count);
    }
}

//Derives Permutation based on orientations in dt. returns an array
//indexed by vertex name containing place in permutation.
int * hyp_DerivePermutation(dt_node * dt,hypergraph * g)
{

  int * Pi = NewArray(int,g->size);
  int * count;
  
  count = New(int);
  (*count) = 0;

  WritePermutation(dt,Pi,count);
  free(count);
  return Pi;
}


//Recursive procedure used by DeriveInvPermutation to record the permutation
//given by an oriented dt into a lookup table
void WriteInvPermutation(dt_node * dt,int * Pi,int * count)
{
  int orientation;

  //If this is a leaf, write it down
  if(dt->v != NOT_LEAF)
    {
      Pi[dt->v] = (*count);
      (*count)++;
    }
  else
    {
      if(dt->orientation == -1) orientation = 0; else orientation = dt->orientation;

      //Otherwise, go to the left child, keeping in mind the orientation,
      //then go to the right child
      WriteInvPermutation(dt->child[orientation], Pi, count);
      WriteInvPermutation(dt->child[1 - orientation], Pi, count);
    }
}

//Derives Permutation based on orientations in dt. returns an array
//indexed by vertex name containing place in permutation.
int * hyp_DeriveInvPermutation(dt_node * dt,hypergraph * g)
{

  int * Pi = NewArray(int,g->size);
  int * count;
  
  count = New(int);
  (*count) = 0;

  WriteInvPermutation(dt,Pi,count);
  free(count);
  return Pi;
}


void hyp_OutputPermutationToFile(char * filename, dt_node * dt, hypergraph * g)
{
  int * pi = hyp_DeriveInvPermutation(dt,g);
  int i;

  int cost = hyp_CostOfOrientation(dt, g);

  FILE * fp;

  fp = fopen(filename, "w");

  if(fp == NULL)
    {
      printf("Cannot Write to File\n");
      return;
    }
  
  printf("Opened File %s\n", filename);


  fprintf(fp,"# Hypergraph Layout File %s\n",filename);
  fprintf(fp,"# Cost:\n");
  fprintf(fp,"%i\n",cost);
  fprintf(fp,"# Nodes\n");
  fprintf(fp,"%i\n",g->size);
  //fprintf(fp,"# Hyperedges\n");
  //fprintf(fp,"%i\n",g->hyperedges);
  fprintf(fp,"# Permutation:\n");
  



  for(i=0; i<g->size; i++)
    {
      fprintf(fp,"%i ",pi[i]);

    }
  fprintf(fp,"\n");

  fprintf(fp,"# End of file\n");

  fclose(fp);

}


//Computes the cost of an orientation setting 
int hyp_CostOfOrientation(dt_node * dt, hypergraph * g)
{

  int count=0;
  int i,j;
  int max,min,current;

  //Returns an array indexed by vertex name containing place in permuation (Really Pi^-1)
  int * Pi = hyp_DeriveInvPermutation(dt,g);
  


#ifdef LINEAR_ARRANGEMENT

  //for each hyperedge e
  for(i=0; i<g->hyperedges; i++)
    {
      //for each node in e
      for(j=0; j<SizeOfHyperedge(i,g); j++)
	{
	  current = Pi[g->vertex_array[g->hyperedge_array[i] + j]];
	  if ((j == 0) || (current > max))
	    max = current;
	  if ((j == 0) || (current < min))
	    min = current;
	}
      count += max - min;
    }
  
#endif

  free(Pi);

#ifdef OVERFLOW_TEST
  if(count > 536870912) printf("OVERFLOW!!!\n");
#endif

  return count;

}


//--------------------------------------------------------------------
//------------------Computing the optimal exhaustively----------------
//--------------------------------------------------------------------


//Initializes an array holding a complete orientation setting.
//Used by ComputeOptimalExhaustively
void hyp_InitOrientationArray(int * orientation_array, int size)
{
  int i;

  for(i=0; i<size; i++)
    orientation_array[i] = 0;

}

//Augments the orientation array to the next orientation
//Used by ComputeOptimalExhaustively
int hyp_AugmentOrientationArray(int * orientation_array, int size)
{
  int i = size-1;
  
  while((orientation_array[i] == 1) && (i > 0))
    {
      orientation_array[i] = 0;
      i--;
    }

  if(orientation_array[i] == 0) 
    orientation_array[i] = 1;
  else 
    return 0;

  return 1;
}

//takes an orientation array and applies it to the dt
//Used by ComputeOptimalExhaustively
hyp_ApplyOrientationArray_r(int * orientation_array, dt_node * dt, int * count) 
{

  if(dt->v != NOT_LEAF)
    return;

  dt->orientation = orientation_array[(*count)];
  (*count)++;

  hyp_ApplyOrientationArray_r(orientation_array, dt->child[0], count);
  hyp_ApplyOrientationArray_r(orientation_array, dt->child[1], count);

}

//takes an orientation array and applies it to the dt (Wrapper)
//Used by ComputeOptimalExhaustively
hyp_ApplyOrientationArray(int * orientation_array, dt_node * dt)
{
  int * count = New(int);
  (*count) = 0;

  hyp_ApplyOrientationArray_r(orientation_array,dt,count);

  free(count);
}

//Computes Optimal orientation by exhaustive search.   This is 
//obviously an exponential-time procedure, so should be used only
//for error-checking on small trees.
hyp_ComputeOptimalExhaustively(dt_node * dt, hypergraph * g)
{
  int * orientation_array = NewArray(int, g->size - 1);
  int i,j;
  int min = -1;
  int max = 0;
  int current = 0;
  int tries = 0;

  hyp_InitOrientationArray(orientation_array, g->size-1);

  do
    {
      
      tries++;
      hyp_ApplyOrientationArray(orientation_array, dt);
      current = hyp_CostOfOrientation(dt, g);
      if( (min == -1) ||  (current < min) ) min = current; 
      if( current > max ) max = current;
    } while(hyp_AugmentOrientationArray(orientation_array, g->size-1));
  
  printf("Optimal: %i\n",min);
  printf("Worst: %i\n",max);
  

  free(orientation_array);
  return;
}




//Returns the sum of the inner cuts
int SumInnerCuts(dt_node * dt)
{
  if (dt == NULL)
    return 0;
  else return
         (SumInnerCuts(dt->child[0])
          + SumInnerCuts(dt->child[1])
          + dt->inner_cut);
}                       


int VerifyInnerCuts(hypergraph * g, dt_node * dt) 
{
  
  int sum = SumInnerCuts(dt);
  printf("Inner Cut Sum: %i, Number of Hyperedges: %i\n",sum,g->hyperedges);

  return 0;
}
 

int OverflowTest()
{
  int i;
  int j=0;

  for(i=1; j<100; i=i*2)
    {
      printf("%i: %i\n",j,i);
      j++;

    }
}
