#include <stdio.h>
#include <math.h>
#include "ot.h"
#include "hyperorient.h"

#define LOWER_PERMUTATION_THRESHOLD 8
#define ITERATIVE_THRESHOLD 10

dt_node * hyp_BuildDtRecursive(dt_node * parent, hypergraph * g, int * pi, int pi_size, int ub_factor)
{
  int i,left,right;
  int cut_size;
  int min,min_index;

  dt_node * new_dt = New(dt_node);
  new_dt->lca_flag = -1;
  new_dt->parent = parent;
  new_dt->orientation = 0;
  new_dt->inner_cut = 0;
  new_dt->inner_cut_temp = 0;
  new_dt->subtree_size = pi_size;


#ifdef DEBUG
  printf("Size of Pi: %i\n",pi_size);
#endif
#ifdef PRINT_OUT_PARTITION
  //output this partition
  printf("{ ");
  for(i=0; i<pi_size; i++)
    printf("%i, ",pi[i]);
  printf("}\n");
#endif





  if(pi_size == 1)
    {
      //build a dt leaf

#ifdef DEBUG
      printf("Leaf %i\n",pi[0]);
#endif

      //Set children to null and assign index as counter
      new_dt->child[0] = NULL;
      new_dt->child[1] = NULL;
      new_dt->v = pi[0];
      g->ref_array[pi[0]] = new_dt;

    }
  else if(pi_size <= LOWER_PERMUTATION_THRESHOLD)
    {
      //divide exactly in half
      new_dt->v = NOT_LEAF;
      
#ifdef DEBUG
      printf("Dividing in half:\n");
#endif
      //first half
      new_dt->child[0] = hyp_BuildDtRecursive(new_dt, 
					  g, 
					  pi, 
					  pi_size / 2, 
					  ub_factor);

      //second half
      new_dt->child[1] = hyp_BuildDtRecursive(new_dt, 
					  g, 
					  &(pi[pi_size/2]), 
					  pi_size - (pi_size / 2), 
					  ub_factor);
    }
  else
    {

      //find min cut within ub_factor % of middle
      new_dt->v = NOT_LEAF;

      //find the left and right sides of range defined by ub_factor
      left = (pi_size/2) - ((ub_factor * pi_size) / 100); 
      right = (pi_size/2) + ((ub_factor * pi_size) / 100); 
      
#ifdef DEBUG
      printf("Left: %i, Right: %i\n",left,right);
#endif


      //don't allow empty sides
      if(left < 1) left = 1;
      if(right > pi_size) right = pi_size;

      //for each cut to the left of node pi[i] (do not consider cuts with empty sides)
      if(left >= right)
	min_index = left;
      else
	{
	 
#ifdef RANDOM_CUT
	  min_index = left + RandomNumber(right - left) - 1;
#endif
	  
	}
#ifdef DEBUG
      printf("Using Min Cut\n");
#endif

      //make the cut at min:
      //first half
      new_dt->child[0] = hyp_BuildDtRecursive(new_dt, 
					  g, 
					  pi, 
					  min_index, 
					  ub_factor);

      //second half
      new_dt->child[1] = hyp_BuildDtRecursive(new_dt, 
					  g, 
					  &(pi[min_index]), 
					  pi_size - min_index, 
					  ub_factor);


    }

  return new_dt;

}

dt_node * hyp_LocalImprove(hypergraph * g, dt_node * dt, int ub_factor, int * score)
{

  int s,t,cost1,cost2;
  int * pi;
  dt_node * dt_new;


  pi = hyp_DerivePermutation(dt,g);

  cost1 = hyp_CostOfOrientation(dt, g);

  dt_new = hyp_BuildDtRecursive(NULL, g, pi, g->size, ub_factor);


#ifdef EVAL_LOCAL_IMPROVE


  cost2 = hyp_CostOfOrientation(dt_new, g);

  printf( "These costs should be equal: %i, %i\n",cost1,cost2);

#endif




  s = time(NULL);
  (*score) = hyp_OrientAndEvaluateOnTheFly(g, dt_new);
  t = time(NULL);

#ifdef EVAL_LOCAL_IMPROVE
  cost2 = hyp_CostOfOrientation(dt_new, g);
  printf( "%i should be more than or equal to %i\n",cost1,cost2);
#endif

  printf("%i, %i\n",(*score),t-s);

  
  return dt_new;  
  
}



dt_node * hyp_LocalImproveIterate(hypergraph * g, dt_node * dt, int ub_factor, int trials, int * min_score)
{
  int s,t,i,counter = 0;
  dt_node *dt1, *dt2;
  int * score = New(int);

  dt1 = dt;

  printf("#trial,score,time to orient\n");

  s = time(NULL);
  i = 0;
  if(trials == -1)
    {
      
      

      while(counter < ITERATIVE_THRESHOLD)
	{
	
	  printf("%i, ",i++);
	  dt2 = hyp_LocalImprove(g, dt1, ub_factor, score);
	  
	  if((*min_score) > (*score)) 
	    {
	      (*min_score) = (*score);
	      counter = 0;
	    }
	  else
	    counter++;
	  
	  
	  FreeDt(dt1);
	  
	  dt1 = dt2;
	  
	}
      
    }
  else
    {
      for(i=1;i<=trials;i++)
	{
	  printf("%i, ",i);
	  

	  dt2 = hyp_LocalImprove(g, dt1, ub_factor, score);
	  
	  if((*min_score) > (*score)) (*min_score) = (*score);
	  
	  FreeDt(dt1);
	  
	  dt1 = dt2;
	  
	}
    }

  free(score);

  t = time(NULL);

  printf("#total time: %i\n",t-s);

  return dt1;

}














